
/*
 *
 * robot.c - The robots
 *
 */

#include <stdio.h>
#include <math.h>

#include "common.h"
#include "random.h"
#include "video.h"
#include "input.h"
#include "staticGeometry.h"
#include "collectables.h"
#include "robots.h"
#include "game.h"
#include "particles.h"
#include "baubles.h"
#include "network.h"

// -----------------------------------------------------------------------
// variables
// -----------------------------------------------------------------------

// all robots
struct robot robots[ROBOTS_MAX];

// the currently selected robot
int robotSelected;

// dust particle timer
float robotsParticleTimerDust = 0.0f;

// the robots health values
int robotsHealthValues[] = {
	-1, // NONE
	20, // MICROWAVE
	8,  // TOASTER
};

// my running robot ID
int robotMyID;

// -----------------------------------------------------------------------
// init the robots
// -----------------------------------------------------------------------

void robotsInit(void) {

	int i;

	for (i = 0; i < ROBOTS_MAX; i++)
		robots[i].type = ROBOT_TYPE_NONE;

	robotSelected = 0;
	robotMyID = 0;
}

// -----------------------------------------------------------------------
// let the robots think
// -----------------------------------------------------------------------

void robotsIteration(float deltaTime) {

	struct robot *r, *rNear;
	float dX, dY, distance, tX, tY, a1, a2, distance1, distance2;
	int i, j, c;

	robotsParticleTimerDust += deltaTime;

	for (i = 0; i < ROBOTS_MAX; i++) {
		r = &robots[i];

		if (r->type == ROBOT_TYPE_NONE)
			continue;

		// explode?
		if ((r->status & ROBOT_STATUS_EXPLODE) != 0) {
			// yes
			r->harvestTimer += deltaTime;

			if (r->harvestTimer > 1000.0f) {
				r->status |= ROBOT_STATUS_INVISIBLE;
				tileMap[baseCoordinates[r->targetBase*2 + 1]][baseCoordinates[r->targetBase*2 + 0]] = TILE_DARK_SAND_1;
				baseEnabled[r->targetBase] = NO;
			}

			// vanish?
			if (r->harvestTimer > 2000.0f)
				r->type = ROBOT_TYPE_NONE;

			continue;
		}

		if (r->owner != networkMyID) {
			// dead reckon the bots that i don't own
			r->rX += r->rXSpeed * deltaTime;
			r->rY += r->rYSpeed * deltaTime;
			r->rZ += r->rZSpeed * deltaTime;
			r->angle += r->angleSpeed * deltaTime;

			// add dust particles
			dX = r->rXSpeed;
			if (dX < 0.0f)
				dX = -dX;
			dY = r->rYSpeed;
			if (dY < 0.0f)
				dY = -dY;

			if (dX > 0.001f || dY > 0.001f)
				robotAddParticlesDust(r, r->rXSpeed, r->rYSpeed);

			continue;
		}

		// start to explode?
		if (r->health <= 0 && (r->status & ROBOT_STATUS_EXPLODE) == 0) {
			// yes
			r->status |= ROBOT_STATUS_EXPLODE;
			r->harvestTimer = 0.0f;

			if (i == robotSelected)
				robotsSelectNext();
		}

		// would i appreciate information?
		if (r->knowsAnInterestingSpot == NO && (r->state == ROBOT_STATE_IDLE || r->state == ROBOT_STATE_GOTO_TARGET)) {
			// is some other robot near me?
			for (j = 0; j < ROBOTS_MAX; j++) {
				rNear = &robots[j];

				if (rNear->type == ROBOT_TYPE_NONE)
					continue;
				if (i == j)
					continue;

				dX = r->rX - rNear->rX;
				dY = r->rY - rNear->rY;
				distance = dX*dX + dY*dY;
				if (distance < 100.0f*100.0f && rNear->knowsAnInterestingSpot == YES) {
					// he knows a good spot! clone it
					r->knowsAnInterestingSpot = YES;
					r->interestingX = rNear->interestingX;
					r->interestingY = rNear->interestingY;
					break;
				}
			}
		}

		// is there something to be harvested near?
		c = -1;
		distance = 0.0f;
		for (j = 0; j < COLLECTABLES_MAX; j++) {
			if (collectables[j].type == COLLECTABLE_TYPE_NONE)
				continue;

			dX = collectables[j].cX - r->rX;
			dY = collectables[j].cY - r->rY;
			distance1 = dX*dX + dY*dY;
			if (c < 0 || distance1 < distance) {
				c = j;
				distance = distance1;
			}
		}

		if (c >= 0 && distance < 100.0f*100.0f) {
			// remember it
			r->interestingX = collectables[c].cX;
			r->interestingY = collectables[c].cY;
			r->knowsAnInterestingSpot = YES;
		}

		if (r->state == ROBOT_STATE_HARVEST) {
			if (c >= 0 && collectables[c].cX == r->harvestX && collectables[c].cY == r->harvestY) {
				r->harvestTimer -= deltaTime;
				if (r->harvestTimer <= 0.0f) {
					// take it!
					r->payload = collectables[c].type;
					collectables[c].type = COLLECTABLE_TYPE_REMOVE;

					if (baseEnabled[r->owner] == YES) {
						// return to base
						r->state = ROBOT_STATE_GOTO_BASE;
						r->targetX = baseCoordinates[r->owner*2 + 0] * tileDX;
						r->targetY = baseCoordinates[r->owner*2 + 1] * tileDY;
						r->targetBase = r->owner;
						r->harvestTimer = ROBOT_DEFAULT_HARVEST_TIME;
					}
					else {
						// no base to return to!
						r->state = ROBOT_STATE_IDLE;
						r->harvestTimer = ROBOT_DEFAULT_HARVEST_TIME;
					}
				}
			}
			else {
				// somebody else took it
				r->state = ROBOT_STATE_IDLE;
				r->harvestTimer = ROBOT_DEFAULT_HARVEST_TIME;
			}
		}

		if (r->type == ROBOT_TYPE_TOASTER) {
			if (c >= 0 && distance < 14.0f*14.0f && r->payload == COLLECTABLE_TYPE_NONE) {
				// harvest!
				r->state = ROBOT_STATE_HARVEST;
				r->harvestX = collectables[c].cX;
				r->harvestY = collectables[c].cY;
			}
		}

		if (r->type == ROBOT_TYPE_MICROWAVE) {
			if (r->state == ROBOT_STATE_IDLE || r->state == ROBOT_STATE_GOTO_TARGET) {
				// see any enemies near?
				for (j = 0; j < ROBOTS_MAX; j++) {
					rNear = &robots[j];

					if (rNear->type == ROBOT_TYPE_NONE)
						continue;
					if (rNear->owner == r->owner)
						continue;
					if (i == j)
						continue;

					dX = r->rX - rNear->rX;
					dY = r->rY - rNear->rY;
					distance = dX*dX + dY*dY;
					if (distance < 100.0f*100.0f) {
						r->state = ROBOT_STATE_ATTACK;
						r->targetRobot = j;
						break;
					}
				}
			}
			else if (r->state == ROBOT_STATE_ATTACK) {
				dX = r->rX - robots[r->targetRobot].rX;
				dY = r->rY - robots[r->targetRobot].rY;
				distance = dX*dX + dY*dY;
				if (distance > 110.0f*110.0f)
					r->state = ROBOT_STATE_IDLE;
				else {
					r->targetX = robots[r->targetRobot].rX;
					r->targetY = robots[r->targetRobot].rY;
				}
			}
		}

		if (r->state == ROBOT_STATE_BASE_1) {
			// go down, to the base
			r->rZ -= deltaTime * 0.05f;
			if (r->rZ < -50.0f) {
				if (r->targetBase == r->owner) {
					// offload the payload
					r->payload = COLLECTABLE_TYPE_NONE;
					// refuel and fix
					r->gas = ROBOT_DEFAULT_GAS;
					r->health = robotsHealthValues[r->type];
				}
				r->state = ROBOT_STATE_BASE_2;
				r->rZ = -50.0f;
			}
		}

		if (r->state == ROBOT_STATE_BASE_2) {
			// enemy base?
			if (r->targetBase != r->owner) {
				r->health = 0;
			}
			else {
				// go up, to the ground
				r->rZ += deltaTime * 0.05f;
				if (r->rZ > 0.0f) {
					r->rZ = 0.0f;

					// go to the last interesting position
					if (r->type == ROBOT_TYPE_TOASTER) {
						r->targetX = r->interestingX;
						r->targetY = r->interestingY;
					}
					else {
						r->targetX = r->guardX;
						r->targetY = r->guardY;
					}
					r->state = ROBOT_STATE_GOTO_TARGET;
				}
			}
		}

		if (r->state == ROBOT_STATE_IDLE) {
			if (r->gas < ROBOT_DEFAULT_GAS * 0.5f && baseEnabled[r->owner] == YES) {
				// return back to base for refuelling
				r->state = ROBOT_STATE_GOTO_BASE;
				r->targetX = baseCoordinates[r->owner*2 + 0] * tileDX;
				r->targetY = baseCoordinates[r->owner*2 + 1] * tileDY;
				r->targetBase = r->owner;
			}
			else if (r->type == ROBOT_TYPE_TOASTER && c >= 0 && distance < 100.0f*100.0f) {
				// got a new target
				r->targetX = collectables[c].cX;
				r->targetY = collectables[c].cY;
				r->state = ROBOT_STATE_GOTO_TARGET;
			}
			else if (r->type == ROBOT_TYPE_TOASTER && r->knowsAnInterestingSpot == YES) {
				// go to the known interesting position
				r->targetX = r->interestingX;
				r->targetY = r->interestingY;
				r->state = ROBOT_STATE_GOTO_TARGET;
			}
			else {
				// no real target, randomize something
				while (1) {
					if (r->type == ROBOT_TYPE_TOASTER) {
						r->targetX = r->rX;
						r->targetY = r->rY;
					}
					else {
						r->targetX = r->guardX;
						r->targetY = r->guardY;
					}

					r->targetX += randomNextInt() % 100;
					r->targetY += randomNextInt() % 100;
					dX = r->rX - r->targetX;
					dY = r->rY - r->targetY;
					if (dX*dX + dY*dY >= 50.0f*50.0f && robotToTileCollision(r->targetX, r->targetY) == NO && robotToBaubleCollision(r->targetX, r->targetY) == NO)
						break;
				}
				r->state = ROBOT_STATE_GOTO_TARGET;

				// it's also our focus of interest
				r->interestingX = r->targetX;
				r->interestingY = r->targetY;
				r->knowsAnInterestingSpot = NO;
			}
		}

		// move towards the target?
		if (r->state == ROBOT_STATE_GOTO_TARGET || r->state == ROBOT_STATE_GOTO_BASE || r->state == ROBOT_STATE_ATTACK) {
			dX = r->targetX - r->rX;
			dY = r->targetY - r->rY;
			distance = dX*dX + dY*dY;
			if (distance < 1.0f*1.0f) {
				// target reached!
				if (r->state == ROBOT_STATE_GOTO_TARGET) {
					if (r->payload == COLLECTABLE_TYPE_NONE) {
						dX = r->interestingX - r->rX;
						dY = r->interestingY - r->rY;
						distance = dX*dX + dY*dY;
						if (distance < 1.0f*1.0f)
							r->knowsAnInterestingSpot = NO;
						r->state = ROBOT_STATE_IDLE;
					}
					else if (r->state == ROBOT_STATE_ATTACK) {


					}
					else if (baseEnabled[r->owner] == YES) {
						r->state = ROBOT_STATE_GOTO_BASE;
						r->targetX = baseCoordinates[r->owner*2 + 0] * tileDX;
						r->targetY = baseCoordinates[r->owner*2 + 1] * tileDY;
						r->targetBase = r->owner;
					}
					else
						r->state = ROBOT_STATE_IDLE;
				}
				else if (r->state == ROBOT_STATE_GOTO_BASE) {
					if (baseEnabled[r->targetBase] == YES) {
						// go into the base
						r->state = ROBOT_STATE_BASE_1;
					}
					else
						r->state = ROBOT_STATE_IDLE;
				}
			}
			else {
				// move towards the target
				a1 = r->angle;

				if (r->turnTimer <= 0.0f) {
					dX = r->targetX - (r->rX + sinf(a1 * M_PI / 180.0f));
					dY = r->targetY - (r->rY + cosf(a1 * M_PI / 180.0f));
					distance1 = dX*dX + dY*dY;

					a2 = r->angle + (deltaTime * (0.2f + ((float)(randomNextInt() % 10))*0.003f));
					dX = r->targetX - (r->rX + sinf(a2 * M_PI / 180.0f));
					dY = r->targetY - (r->rY + cosf(a2 * M_PI / 180.0f));
					distance2 = dX*dX + dY*dY;

					if (distance2 < distance1) {
						distance1 = distance2;
						a1 = a2;
					}

					a2 = r->angle - (deltaTime * (0.2f + ((float)(randomNextInt() % 10))*0.003f));
					dX = r->targetX - (r->rX + sinf(a2 * M_PI / 180.0f));
					dY = r->targetY - (r->rY + cosf(a2 * M_PI / 180.0f));
					distance2 = dX*dX + dY*dY;

					if (distance2 < distance1)
						a1 = a2;
				}
				else
					r->turnTimer -= deltaTime;

				dX = sinf(a1 * M_PI / 180.0f);
				dY = cosf(a1 * M_PI / 180.0f);
				distance = sqrtf(dX*dX + dY*dY);
				distance = (deltaTime * 0.02f) / distance;

				// cargo slows us down a little
				if (r->payload != COLLECTABLE_TYPE_NONE)
					distance *= 0.7f;

				// out of gas we move very slow
				if (r->gas <= 0.0f)
					distance *= 0.2f;

				dX *= distance;
				dY *= distance;

				// TODO: add external forces
				dX += r->externalX;
				dY += r->externalY;
				r->externalX = 0.0f;
				r->externalY = 0.0f;

				tX = r->rX + dX;
				tY = r->rY + dY;

				// add dust particles
				robotAddParticlesDust(r, dX / deltaTime, dY / deltaTime);

				r->angle = a1;

				// if we back off, in case of collisions...
				a1 = r->rX - dX;
				a2 = r->rY - dY;

				if (r->rX < r->targetX && tX > r->targetX)
					tX = r->targetX;
				else if (r->rX > r->targetX && tX < r->targetX)
					tX = r->targetX;

				if (r->rY < r->targetY && tY > r->targetY)
					tY = r->targetY;
				else if (r->rY > r->targetY && tY < r->targetY)
					tY = r->targetY;

				// check collisions
				if (robotToRobotCollision(i, r->rX, tY) == YES || robotToTileCollision(r->rX, tY) == YES || robotToBaubleCollision(r->rX, tY) == YES) {
					r->angle -= deltaTime * 0.1f;
					r->turnTimer = 300.0f,
					tY = r->rY;
				}
				if (robotToRobotCollision(i, tX, r->rY) == YES || robotToTileCollision(tX, r->rY) == YES || robotToBaubleCollision(tX, r->rY) == YES) {
					if (r->turnTimer < 300.0f) {
						r->angle -= deltaTime * 0.1f;
						r->turnTimer = 300.0f;
					}
					tX = r->rX;
				}

				/*
				if (tX == r->rX && tY == r->rY) {
					// couldn't move at all -> back off
					tX = a1;
					tY = a2;

					if (robotToRobotCollision(i, r->rX, tY) == YES || robotToTileCollision(r->rX, tY) == YES || robotToBaubleCollision(r->rX, tY) == YES)
						tY = r->rY;
					if (robotToRobotCollision(i, tX, r->rY) == YES || robotToTileCollision(tX, r->rY) == YES || robotToBaubleCollision(tX, r->rY) == YES)
						tX = r->rX;
				}
				*/

				if (tX != r->rX || tY != r->rY) {
					// consume gas
					r->gas -= deltaTime;
				}

				r->rX = tX;
				r->rY = tY;
			}
		}
	}

	while (robotsParticleTimerDust > 100.0f)
		robotsParticleTimerDust -= 100.0f;
}

// -----------------------------------------------------------------------
// add dust particles
// -----------------------------------------------------------------------

void robotAddParticlesDust(struct robot *r, float dX, float dY) {

	float pX, pY;

	if (robotsParticleTimerDust > 100.0f) {
		pX = r->rX;
		pY = r->rY;
		pX += (randomNextInt() % 35) * 0.1f;
		pY += (randomNextInt() % 35) * 0.1f;
		particlesAdd(PARTICLE_TYPE_DUST, pX, pY, r->rZ - 1550.0f, -dX, -dY, 0.0f);
	}
}

// -----------------------------------------------------------------------
// draw the robots
// -----------------------------------------------------------------------

void robotsDraw(void) {

	float dX;
	int i, j;

	for (i = 0; i < ROBOTS_MAX; i++) {
		if (robots[i].type == ROBOT_TYPE_NONE)
			continue;

		videoSetColor(255, 255, 255, 255);

		// explode?
		if ((robots[i].status & ROBOT_STATUS_EXPLODE) != 0) {
			if (robots[i].harvestTimer > 1000.0f)
				dX = 1000.0f - (robots[i].harvestTimer - 1000.0f);
			else
				dX = robots[i].harvestTimer;

			if (robots[i].state == ROBOT_STATE_BASE_2)
				dX /= 150.0f;
			else
				dX /= 500.0f;

			if (dX < 0.0f)
				dX = 0.0f;

			videoMatrixPush();
			videoTranslate3f(robots[i].rX, robots[i].rY, robots[i].rZ - 1549.0f);
			videoRotate3f(0.0f, 0.0f, -runTime * 0.1f);
			videoScale3f(dX, dX, dX);
			videoDrawMeshTextured(verticesBarrel, verticesBarrelTriangles);
			videoMatrixPop();
		}

		if ((robots[i].status & ROBOT_STATUS_INVISIBLE) != 0)
			continue;

		// draw the robot
		videoMatrixPush();
		videoTranslate3f(robots[i].rX, robots[i].rY, robots[i].rZ - 1549.0f);
		videoRotate3f(0.0f, 0.0f, -robots[i].angle - 180.0f);
		if (robots[i].type == ROBOT_TYPE_MICROWAVE)
			videoDrawMeshTextured(verticesBotMicrowave, verticesBotMicrowaveTriangles);
		else
			videoDrawMeshTextured(verticesBotToaster, verticesBotToasterTriangles);
		videoMatrixPop();

		// draw the owner marker?
		if (robots[i].owner != networkMyID) {
			j = robots[i].owner << 2;
			videoSetColor(gamePlayerColors[j+0], gamePlayerColors[j+1], gamePlayerColors[j+2], 255);
			videoMatrixPush();
			videoTranslate3f(robots[i].rX, robots[i].rY, robots[i].rZ - 1556.0f);
			videoDrawMeshPlain(verticesCursor1, verticesCursor1Triangles);
			videoMatrixPop();
		}
	}

	videoSetColor(255, 0, 0, 255);

	for (i = 0; i < ROBOTS_MAX; i++) {
		if (robots[i].type == ROBOT_TYPE_NONE)
			continue;
		if ((robots[i].status & ROBOT_STATUS_INVISIBLE) != 0)
			continue;
		if (robots[i].owner != networkMyID)
			continue;

		dX = 20.0f * robots[i].health / robotsHealthValues[robots[i].type];

		// draw the health bar
		videoMatrixPush();
		videoTranslate3f(robots[i].rX, robots[i].rY - 2.0f, robots[i].rZ - 1549.0f + 15.3f);
		videoRotate3f(-gameCameraAngle, 0.0f, 0.0f);
		videoDrawRectangle2D(dX, 2.0f);
		videoMatrixPop();
	}

	videoSetColor(55, 147, 186, 255);

	for (i = 0; i < ROBOTS_MAX; i++) {
		if (robots[i].type == ROBOT_TYPE_NONE)
			continue;
		if ((robots[i].status & ROBOT_STATUS_INVISIBLE) != 0)
			continue;
		if (robots[i].owner != networkMyID)
			continue;

		dX = 20.0f * robots[i].gas / ROBOT_DEFAULT_GAS;

		// draw the gas bar
		videoMatrixPush();
		videoTranslate3f(robots[i].rX, robots[i].rY, robots[i].rZ - 1549.0f + 13.0f);
		videoRotate3f(-gameCameraAngle, 0.0f, 0.0f);
		videoDrawRectangle2D(dX, 2.0f);
		videoMatrixPop();
	}

	i = networkMyID << 2;
	videoSetColor(gamePlayerColors[i+0], gamePlayerColors[i+1], gamePlayerColors[i+2], 255);

	// draw the selection marker?
	if (robotSelected >= 0) {
		videoMatrixPush();

		videoTranslate3f(robots[robotSelected].rX, robots[robotSelected].rY, robots[robotSelected].rZ - 1555.0f);
		videoRotate3f(0.0f, 0.0f, runTime * 0.1f);
		videoDrawMeshPlain(verticesCursor1, verticesCursor1Triangles);

		videoMatrixPop();
	}

	// draw the target marker?
	if (robotSelected >= 0) {
		videoMatrixPush();

		videoTranslate3f(robots[robotSelected].targetX, robots[robotSelected].targetY, -1555.0f);
		videoRotate3f(0.0f, 0.0f, runTime * 0.1f);
		videoDrawMeshPlain(verticesCursor2, verticesCursor2Triangles);

		videoMatrixPop();
	}

	i = networkMyID << 2;
	videoSetColor(gamePlayerColors[i+0] >> 1, gamePlayerColors[i+1] >> 1, gamePlayerColors[i+2] >> 1, 255);

	// draw the target pointer?
	if (robotSelected >= 0) {
		videoMatrixPush();

		videoTranslate3f(gameCameraX, gameCameraY + gameCameraAngle*(-1500.0f - gameCameraZ)*0.017f - 50.0f, -1555.2f);
		videoScale3f(0.8f, 0.8f, 1.0f);
		videoDrawMeshPlain(verticesCursor2, verticesCursor2Triangles);

		videoMatrixPop();
	}

	videoSetColor(255, 255, 255, 255);
}

// -----------------------------------------------------------------------
// select the next robot
// -----------------------------------------------------------------------

void robotsSelectNext(void) {

	int i, j;

	j = robotSelected;
	for (i = 0; i < ROBOTS_MAX; i++) {
		j++;
		if (j == ROBOTS_MAX)
			j = 0;
		if (robots[j].type != ROBOT_TYPE_NONE && robots[j].owner == networkMyID && (robots[j].status & ROBOT_STATUS_EXPLODE) == 0) {
			robotSelected = j;
			return;
		}
	}
}

// -----------------------------------------------------------------------
// does a robot collide with a bauble?
// -----------------------------------------------------------------------

int  robotToBaubleCollision(float nX, float nY) {

	float dX, dY, r;
	int i;

	for (i = 0; i < BAUBLES_MAX; i++) {
		if (baubles[i].type == BAUBLE_TYPE_NONE)
			continue;

		dX = nX - baubles[i].cX;
		dY = nY - baubles[i].cY;
		r = baubles[i].scale * 10.0f;

		if (dX*dX + dY*dY < r*r)
			return YES;
	}

	return NO;
}

// -----------------------------------------------------------------------
// does a robot collide with a tile?
// -----------------------------------------------------------------------

int  robotToTileCollision(float nX, float nY) {

	int tY, tX;

	nX += tileDX * 0.5f;
	nY += tileDY * 0.5f;

	tX = (int)(nX / tileDX);
	tY = (int)(nY / tileDY);
	if (tileCollisions[tileMap[tY][tX]] == YES)
		return YES;

	return NO;
}

// -----------------------------------------------------------------------
// does the specified robot collide with any other robot?
// -----------------------------------------------------------------------

int  robotToRobotCollision(int robot, float nX, float nY) {

	float dX, dY, dZ, distance;
	int i;

	for (i = 0; i < ROBOTS_MAX; i++) {
		if (robots[i].type == ROBOT_TYPE_NONE)
			continue;
		if (i == robot)
			continue;

		dX = nX - robots[i].rX;
		dY = nY - robots[i].rY;
		dZ = robots[robot].rZ - robots[i].rZ;
		distance = dX*dX + dY*dY + dZ*dZ;
		if (distance < 15.0f*15.0f) {
			// are we going closer?
			dX = robots[robot].rX - robots[i].rX;
			dY = robots[robot].rY - robots[i].rY;
			if (distance < dX*dX + dY*dY + dZ*dZ)
				return YES;
		}
	}

	return NO;
}

// -----------------------------------------------------------------------
// update robot information (got from the network)
// -----------------------------------------------------------------------

void robotsUpdate(int senderID, int robotID, int type, float rX, float rY, float rZ, float angle, int packetNumber) {

	float dT;
	int i;

	//fprintf(stderr, "senderID = %d robotID = %d type = %d rX = %.2f rY = %.2f rZ = %.2f\n", senderID, robotID, type, rX, rY, rZ);

	for (i = 0; i < ROBOTS_MAX; i++) {
		if (robots[i].type == ROBOT_TYPE_NONE)
			continue;
		if (robots[i].owner != senderID || robots[i].iD != robotID)
			continue;

		// found it

		// dead reckon position and angle
		dT = runTime - robots[i].updatePacketTime;
		if (dT < 0.0001) {
			robots[i].rXSpeed = 0.0f;
			robots[i].rYSpeed = 0.0f;
			robots[i].rZSpeed = 0.0f;
			robots[i].angleSpeed = 0.0f;
		}
		else {
			dT = 1.0f / dT;
			robots[i].rXSpeed = (rX - robots[i].rXStart) * dT;
			robots[i].rYSpeed = (rY - robots[i].rYStart) * dT;
			robots[i].rZSpeed = (rZ - robots[i].rZStart) * dT;
			robots[i].angleSpeed = (angle - robots[i].angleStart) * dT;
		}

		// update
		robots[i].rX = rX;
		robots[i].rY = rY;
		robots[i].rZ = rZ;
		robots[i].angle = angle;
		robots[i].rXStart = rX;
		robots[i].rYStart = rY;
		robots[i].rZStart = rZ;
		robots[i].angleStart = angle;
		robots[i].updatePacketNumber = packetNumber;
		robots[i].updatePacketTime = runTime;

		return;
	}

	// we didn't find the robot -> add it!
	i = robotsAdd(type, rX, rY, angle, senderID, robotID, ROBOT_STATE_IDLE);

	// update the packet number
	if (i >= 0)
		robots[i].updatePacketNumber = packetNumber;
}

// -----------------------------------------------------------------------
// remove the robots that were not updated
// -----------------------------------------------------------------------

void robotsRemoveMissing(int senderID, int packetNumber) {

	int i;

	for (i = 0; i < ROBOTS_MAX; i++) {
		if (robots[i].type == ROBOT_TYPE_NONE)
			continue;
		if (robots[i].owner != senderID)
			continue;
		if (robots[i].updatePacketNumber == packetNumber)
			continue;

		// no update -> remove the robot
		robots[i].type = ROBOT_TYPE_NONE;
	}
}

// -----------------------------------------------------------------------
// add a robot
// -----------------------------------------------------------------------

int  robotsAdd(int type, float rX, float rY, float angle, int owner, int iD, int state) {

	int i;

	for (i = 0; i < ROBOTS_MAX; i++) {
		if (robots[i].type == ROBOT_TYPE_NONE) {
			robots[i].type = type;
			robots[i].status = 0;
			robots[i].rX = rX;
			robots[i].rY = rY;
			robots[i].rZ = 0.0f;
			robots[i].rXStart = rX;
			robots[i].rYStart = rY;
			robots[i].rZStart = 0.0f;
			robots[i].rXSpeed = 0.0f;
			robots[i].rYSpeed = 0.0f;
			robots[i].rZSpeed = 0.0f;
			robots[i].externalX = 0.0f;
			robots[i].externalY = 0.0f;
			robots[i].turnTimer = 0.0f;
			robots[i].guardX = rX;
			robots[i].guardY = rY;
			robots[i].angle = angle;
			robots[i].angleStart = angle;
			robots[i].angleSpeed = 0.0f;
			robots[i].owner = owner;
			robots[i].state = state;
			robots[i].payload = COLLECTABLE_TYPE_NONE;
			robots[i].gas = ROBOT_DEFAULT_GAS;
			robots[i].health = robotsHealthValues[type];
			robots[i].harvestTimer = ROBOT_DEFAULT_HARVEST_TIME;
			robots[i].knowsAnInterestingSpot = NO;
			robots[i].updatePacketNumber = -1;
			robots[i].updatePacketTime = runTime;

			if (iD < 0)
				robots[i].iD = robotMyID++;
			else
				robots[i].iD = iD;

			return i;
		}
	}

	return -1;
}
