
/*
 *
 * network.c - The Posix network subsystem
 *
 */

#define WIN32_LEAN_AND_MEAN		// exclude rarely used stuff from Windows headers

#ifdef WIN32
#include <windows.h>
#include <stdlib.h>
#include <malloc.h>
#include <memory.h>
#include <tchar.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#endif

#ifdef LINUX
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <time.h>
#endif

#ifdef PSP
#include <pspkernel.h>
#include <pspdebug.h>
#include <pspsdk.h>
#include <stdlib.h>
#include <string.h>
#include <pspnet.h>
#include <pspnet_inet.h>
#include <pspnet_apctl.h>
#include <pspnet_resolver.h>
#include <psputility_netmodules.h>
#include <psputility_netparam.h>
#include <pspwlan.h>
#include <errno.h>
#endif

#ifdef WIN32
#include <winsock2.h>
#else
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif

#include "common.h"
#include "network.h"
#include "game.h"
#include "random.h"
#include "robots.h"
#include "collectables.h"

// -----------------------------------------------------------------------
// variables
// -----------------------------------------------------------------------

// server port
unsigned int networkServerPort;

// my port
unsigned int networkClientPort;

// server connection status
int networkServerConnectionStatus;

// the packet numbers
int networkPacketNumberOut;
int networkPacketNumberIn;

// address data structures
struct sockaddr_in networkServerAddress;
struct sockaddr_in networkClientAddress;

// I/O socket
#ifdef WIN32
SOCKET networkSocketFD;
#else
int networkSocketFD;
#endif

// network buffer for data I/O
unsigned char networkBuffer[8*1024];

// server timeout (in seconds)
int networkServerTimeout;

// the last time we got a ping from the server
int networkServerPingTime;

// no response timer
int networkNoResponseTimer;

// my ID in the game
int networkMyID;

// the network send timer
float networkSendTimer;

// -----------------------------------------------------------------------
// init
// -----------------------------------------------------------------------

void networkInit(void) {

#ifdef WIN32
  WSADATA wsaData;

  WSAStartup(MAKEWORD(2,0), &wsaData);
#endif

  networkServerConnectionStatus = NETWORK_SERVER_NO_CONNECTION;
  networkSocketFD = -1;
  networkPacketNumberIn = 0;
  networkPacketNumberOut = 0;

  // init the client address structure
  networkClientAddress.sin_family = AF_INET;

  // reset the server port
  networkServerPort = 9999;

	// use the no response timer later
	networkNoResponseTimer = 0;

	// zero the send timer
	networkSendTimer = 0.0f;

#ifdef PSP
	if (pspSdkLoadInetModules() < 0)
		return;
#endif
}

// -----------------------------------------------------------------------
// send a packet to the server
// -----------------------------------------------------------------------

static void networkServerSend(unsigned char *buffer, int size) {

  // embed the message size into the message
  buffer[0] = size >> 8;
  buffer[1] = size & 0xFF;

  sendto(networkSocketFD, buffer, size, 0, (struct sockaddr *)&networkServerAddress, sizeof(struct sockaddr));

	networkPacketNumberOut++;
}

// -----------------------------------------------------------------------
// send status to the server
// -----------------------------------------------------------------------

void networkServerSendUpdate(float deltaTime) {

	unsigned char *b;
	int i, j, r, t;

	if (networkServerConnectionStatus != NETWORK_SERVER_GAME)
		return;

	networkSendTimer += deltaTime;

	if (networkSendTimer < 100.0f)
		return;

	networkSendTimer -= 100.0f;
	b = networkBuffer;

	// construct an update packet
	i = 2;
	b[i++] = 'U';
	b[i++] = networkMyID;

	// count my robots
	r = 0;
	for (j = 0; j < ROBOTS_MAX; j++) {
		if (robots[j].type == ROBOT_TYPE_NONE)
			continue;
		if (robots[j].owner == networkMyID)
			r++;
	}

	b[i++] = r;

	// send the robots
	for (j = 0; j < ROBOTS_MAX; j++) {
		if (robots[j].type == ROBOT_TYPE_NONE)
			continue;
		if (robots[j].owner != networkMyID)
			continue;

		b[i++] = robots[j].iD;
		b[i++] = robots[j].type;

		t = (int)(robots[j].rX * 256.0f);
		b[i++] = t >> 24;
		b[i++] = t >> 16;
		b[i++] = t >> 8;
		b[i++] = t;

		t = (int)(robots[j].rY * 256.0f);
		b[i++] = t >> 24;
		b[i++] = t >> 16;
		b[i++] = t >> 8;
		b[i++] = t;

		t = (int)(robots[j].rZ * 256.0f);
		b[i++] = t >> 24;
		b[i++] = t >> 16;
		b[i++] = t >> 8;
		b[i++] = t;

		t = (int)(robots[j].angle * 256.0f);
		b[i++] = t >> 24;
		b[i++] = t >> 16;
		b[i++] = t >> 8;
		b[i++] = t;
	}

	// count the collectables
	r = 0;
	for (j = 0; j < COLLECTABLES_MAX; j++) {
		if (collectables[j].type == COLLECTABLE_TYPE_REMOVE)
			r++;
	}

	b[i++] = r;

	// send the collectables
	for (j = 0; j < COLLECTABLES_MAX; j++) {
		if (collectables[j].type != COLLECTABLE_TYPE_REMOVE)
			continue;

		t = (int)(collectables[j].cX * 256.0f);
		b[i++] = t >> 24;
		b[i++] = t >> 16;
		b[i++] = t >> 8;
		b[i++] = t;

		t = (int)(collectables[j].cY * 256.0f);
		b[i++] = t >> 24;
		b[i++] = t >> 16;
		b[i++] = t >> 8;
		b[i++] = t;

		collectables[j].type = COLLECTABLE_TYPE_NONE;
	}

  networkServerSend(b, i);
}

// -----------------------------------------------------------------------
// connect to the network
// -----------------------------------------------------------------------

#ifdef PSP
int networkPSPStatus = 0;
#endif

void networkConnect(void) {

#ifdef PSP
	int err, state;

	if (networkPSPStatus == 3)
		return;

	err = pspSdkInetInit();
	if (err != 0)
		return;

	networkPSPStatus = 1;

	// connect to the first AP
	err = sceNetApctlConnect(0);
	if (err != 0)
		return;

	networkPSPStatus = 2;

	// wait until we have a connection
	while (1) {
		err = sceNetApctlGetState(&state);
		if (err != 0)
			break;

		// connected?
		if (state >= 4)
			break;

		// not trying?
		if (state < 0)
			break;

		// wait 50ms before polling again
		sceKernelDelayThread(50*1000);
	}

	networkPSPStatus = 3;
#endif
}

// -----------------------------------------------------------------------
// disconnect from the network
// -----------------------------------------------------------------------

void networkDisconnect(void) {

}

// -----------------------------------------------------------------------
// disconnect from the server
// -----------------------------------------------------------------------

void networkServerDisconnect(void) {

	if (networkSocketFD >= 0) {
		shutdown(networkSocketFD, 0);

#ifdef WIN32
		closesocket(networkSocketFD);
#else
		close(networkSocketFD);
#endif

		networkSocketFD = -1;
	}

  networkServerConnectionStatus = NETWORK_SERVER_NO_CONNECTION;
  networkPacketNumberIn = 0;
  networkPacketNumberOut = 0;
}

// -----------------------------------------------------------------------
// connect to the game server
// -----------------------------------------------------------------------

void networkServerConnect(char *playerName, char *networkServerName) {

  struct hostent *serverInfo;
  struct sockaddr_in addrMy;
	int i, j;

	// create a socket
	if ((networkSocketFD = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
		return;

	// set the socket to be non-blocking
#ifdef WIN32
	j = 1;
	ioctlsocket(networkSocketFD, FIONBIO, &j);
#endif

	// reset my port
	networkClientPort = 10000 + (randomNextInt() & 31);

	// bind my socket
	addrMy.sin_family = AF_INET;
	addrMy.sin_port = htons(networkClientPort);
	addrMy.sin_addr.s_addr = INADDR_ANY;        // automatically fill with my IP

	bind(networkSocketFD, (struct sockaddr *)&addrMy, sizeof(struct sockaddr));

	if ((serverInfo = gethostbyname(networkServerName)) == NULL)
		return;

	// fill in the server information
	networkServerAddress.sin_family = AF_INET;
	networkServerAddress.sin_port = htons(networkServerPort);
	networkServerAddress.sin_addr.s_addr = *((unsigned long *)(serverInfo->h_addr));

  // send a join
	j = 2;
	networkBuffer[j++] = 'J';

	i = 0;
	while (playerName[i] != 0)
		networkBuffer[j++] = playerName[i++];
	networkBuffer[j++] = 0;

  networkServerSend(networkBuffer, j);

	// if we don't get an answer in 1.5 seconds, we'll display NO RESPONSE
	networkNoResponseTimer = 1500;
}

// -----------------------------------------------------------------------
// send start!
// -----------------------------------------------------------------------

void networkServerSendStart(void) {

	networkBuffer[2] = 'B';
  networkServerSend(networkBuffer, 3);
}

// -----------------------------------------------------------------------
// read the input from network
// -----------------------------------------------------------------------

void networkRead(void) {

	struct sockaddr_in inputAddress;
  int i, size, j, k, m, type;
	float rX, rY, rZ, angle;
	socklen_t len;

  // do we have a connection to the network?
  if (networkSocketFD < 0)
    return;

  // get all the messages
  while (1) {
    len = sizeof(struct sockaddr);
#ifdef WIN32
    i = recvfrom(networkSocketFD, networkBuffer, 8*1024, 0, (struct sockaddr *)&inputAddress, &len);
#else
    i = recvfrom(networkSocketFD, networkBuffer, 8*1024, MSG_DONTWAIT, (struct sockaddr *)&inputAddress, &len);
#endif

    // nothing in the queue?
    if (i < 3)
      return;

		networkPacketNumberIn++;

    // parse the payload size
    size = (networkBuffer[0] << 8) | networkBuffer[1];

    // is the packet ok?
    if (size != i)
			return;

    i = 2;

    // parse the payload
    while (i < size) {
      switch (networkBuffer[i++]) {
			case 'D':
				// DEATH OF A CLIENT

				// client ID
				j = networkBuffer[i++];

				// destroy all the robots
				for (k = 0; k < ROBOTS_MAX; k++) {
					if (robots[k].type == ROBOT_TYPE_NONE)
						continue;
					if (robots[k].owner != j)
						continue;

					robots[k].health = 0;
					robots[k].status |= ROBOT_STATUS_EXPLODE;
					robots[k].harvestTimer = 0.0f;
				}

				break;

			case 'U':
				// UPDATE

				// sender ID
				j = networkBuffer[i++];

				// the number of robots
				k = networkBuffer[i++];

				while (k > 0) {
					// robot ID
					m = networkBuffer[i++];

					// type
					type = networkBuffer[i++];

					rX = ((networkBuffer[i+0] << 24) | (networkBuffer[i+1] << 16) | (networkBuffer[i+2] << 8) | networkBuffer[i+3]) * (1.0f / 256.0f);
					i += 4;
					rY = ((networkBuffer[i+0] << 24) | (networkBuffer[i+1] << 16) | (networkBuffer[i+2] << 8) | networkBuffer[i+3]) * (1.0f / 256.0f);
					i += 4;
					rZ = ((networkBuffer[i+0] << 24) | (networkBuffer[i+1] << 16) | (networkBuffer[i+2] << 8) | networkBuffer[i+3]) * (1.0f / 256.0f);
					i += 4;
					angle = ((networkBuffer[i+0] << 24) | (networkBuffer[i+1] << 16) | (networkBuffer[i+2] << 8) | networkBuffer[i+3]) * (1.0f / 256.0f);
					i += 4;

					robotsUpdate(j, m, type, rX, rY, rZ, angle, networkPacketNumberIn);

					k--;
				}

				robotsRemoveMissing(j, networkPacketNumberIn);

				// the number of collectables that we need to remove
				k = networkBuffer[i++];

				while (k > 0) {
					rX = ((networkBuffer[i+0] << 24) | (networkBuffer[i+1] << 16) | (networkBuffer[i+2] << 8) | networkBuffer[i+3]) * (1.0f / 256.0f);
					i += 4;
					rY = ((networkBuffer[i+0] << 24) | (networkBuffer[i+1] << 16) | (networkBuffer[i+2] << 8) | networkBuffer[i+3]) * (1.0f / 256.0f);
					i += 4;

					collectablesRemove(rX, rY);

					k--;
				}

				break;

			case 'B':
				// BEGIN THE MATCH

				networkServerConnectionStatus = NETWORK_SERVER_GAME;

				break;

			case 'p':
				// PING

				//fprintf(stderr, "p\n");

				// ping back
				networkBuffer[2] = 'p';
				networkServerSend(networkBuffer, 3);

				break;

			case 'N':
				// NAME LIST

				//fprintf(stderr, "N\n");

				j = networkBuffer[i++];

				// reset the name list
				for (k = 0; k < j; k++) {
					m = 0;
					while (networkBuffer[i] != 0)
						playerNameList[k][m++] = networkBuffer[i++];
					i++;
					playerNameList[k][m] = 0;
					playerNameListN[k] = m;
				}

				break;

      case 'W':
				// WELCOME

				//fprintf(stderr, "w\n");

				// reset the no response timer
				networkNoResponseTimer = 0;

				// get my ID
				networkMyID = networkBuffer[i++];

				// update the server status
				networkServerConnectionStatus = NETWORK_SERVER_CONNECTION;

				break;
			}
		}
	}
}
